﻿using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;

namespace OptimisationManagerExtention.ViewExtention
{
    /// <summary>
    /// Данный класс был заимстрован с форума https://stackoverflow.com:
    /// https://stackoverflow.com/questions/2643545/wpf-mvvm-how-to-bind-gridviewcolumn-to-viewmodel-collection
    /// Автор кода: Thomas Levesque
    /// https://stackoverflow.com/users/98713/thomas-levesque
    /// </summary>
    static class GridViewColumns
    {
        #region ColumnsSource
        /// <summary>
        /// Getter для ColumnsSourceProperty
        /// </summary>
        /// <param name="obj">Эллемент управления</param>
        /// <returns></returns>
        [AttachedPropertyBrowsableForType(typeof(GridView))]
        public static object GetColumnsSource(DependencyObject obj)
        {
            return (object)obj.GetValue(ColumnsSourceProperty);
        }
        /// <summary>
        /// Setter ColumnsSourceProperty
        /// </summary>
        /// <param name="obj">Эллемент управления</param>
        /// <param name="value">передаваемое значение</param>
        public static void SetColumnsSource(DependencyObject obj, object value)
        {
            obj.SetValue(ColumnsSourceProperty, value);
        }

        // Using a DependencyProperty as the backing store for ColumnsSource. //--- Thomas Levesque ---//
        // This enables animation, styling, binding, etc... //--- Thomas Levesque ---//
        public static readonly DependencyProperty ColumnsSourceProperty =
            DependencyProperty.RegisterAttached(
                "ColumnsSource",
                typeof(object),
                typeof(GridViewColumns),
                new UIPropertyMetadata(
                    null,
                    ColumnsSourceChanged));
        /// <summary>
        /// Коллбек на событие изменения ColumnsSourceProperty
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="e"></param>
        private static void ColumnsSourceChanged(DependencyObject obj, DependencyPropertyChangedEventArgs e)
        {
            // Приводится передаваемый объект к типу GridView
            GridView gridView = obj as GridView;
            if (gridView != null)
            {
                // Производится отчистка всех колонок
                gridView.Columns.Clear();

                // Если есть станые данные - то они отписываются от коллбека наблюдения за коллекцией колонок
                if (e.OldValue != null)
                {
                    ICollectionView view = CollectionViewSource.GetDefaultView(e.OldValue);
                    if (view != null)
                        RemoveHandlers(gridView, view);
                }

                // Если есть новые данные - то они подписываются на коллбек и
                // на их основе создаются новые колонки
                if (e.NewValue != null)
                {
                    ICollectionView view = CollectionViewSource.GetDefaultView(e.NewValue);
                    if (view != null)
                    {
                        AddHandlers(gridView, view);
                        CreateColumns(gridView, view.Cast<ColumnDescriptor>());
                    }
                }
            }
        }

        #endregion

        #region HeaderTextMember
        /// <summary>
        /// Getter HeaderTextMemberProperty
        /// </summary>
        /// <param name="obj">Элемент управления</param>
        /// <returns>Наименование заголовка столбца</returns>
        [AttachedPropertyBrowsableForType(typeof(GridView))]
        public static string GetHeaderTextMember(DependencyObject obj)
        {
            return (string)obj.GetValue(HeaderTextMemberProperty);
        }
        /// <summary>
        /// Setter HeaderTextMemberProperty
        /// </summary>
        /// <param name="obj">Элемент управления</param>
        /// <param name="value">Значение</param>
        public static void SetHeaderTextMember(DependencyObject obj, string value)
        {
            obj.SetValue(HeaderTextMemberProperty, value);
        }

        // Using a DependencyProperty as the backing store for HeaderTextMember.  //--- Thomas Levesque ---//
        // This enables animation, styling, binding, etc... //--- Thomas Levesque ---//
        public static readonly DependencyProperty HeaderTextMemberProperty =
            DependencyProperty.RegisterAttached("HeaderTextMember", typeof(string), typeof(GridViewColumns), new UIPropertyMetadata(null));
        #endregion

        #region DisplayMemberMember
        /// <summary>
        /// Setter DisplayMemberMemberProperty
        /// </summary>
        /// <param name="obj">Элемент управления</param>
        /// <returns></returns>
        [AttachedPropertyBrowsableForType(typeof(GridView))]
        public static string GetDisplayMemberMember(DependencyObject obj)
        {
            return (string)obj.GetValue(DisplayMemberMemberProperty);
        }
        /// <summary>
        /// Getter DisplayMemberMemberProperty
        /// </summary>
        /// <param name="obj">Элемент управления</param>
        /// <param name="value">Значение</param>
        public static void SetDisplayMemberMember(DependencyObject obj, string value)
        {
            obj.SetValue(DisplayMemberMemberProperty, value);
        }

        // Using a DependencyProperty as the backing store for DisplayMember.  //--- Thomas Levesque ---//
        // This enables animation, styling, binding, etc... //--- Thomas Levesque ---//
        public static readonly DependencyProperty DisplayMemberMemberProperty =
            DependencyProperty.RegisterAttached("DisplayMemberMember", typeof(string), typeof(GridViewColumns), new UIPropertyMetadata(null));
        #endregion

        /// <summary>
        /// Словарь где:
        /// key - коллекция столбцов
        /// value - коллекция всех GridView всех таблиц что исспользуют данное расширение 
        /// с коллекцией столбцов переданной в качестве ключа
        /// </summary>
        private static IDictionary<ICollectionView, List<GridView>> _gridViewsByColumnsSource =
            new Dictionary<ICollectionView, List<GridView>>();

        /// <summary>
        /// Метод выбирающий GridView из словаря "_gridViewsByColumnsSource"
        /// </summary>
        /// <param name="columnSource">Коллекция колонок - ключ</param>
        /// <returns></returns>
        private static List<GridView> GetGridViewsForColumnSource(ICollectionView columnSource)
        {
            List<GridView> gridViews;
            if (!_gridViewsByColumnsSource.TryGetValue(columnSource, out gridViews))
            {
                gridViews = new List<GridView>();
                _gridViewsByColumnsSource.Add(columnSource, gridViews);
            }
            return gridViews;
        }
        /// <summary>
        /// Метод подписывающийся на событие изменения колонок и
        /// Добавляющий новый GrigView в коллекцию "_gridViewsByColumnsSource"
        /// </summary>
        /// <param name="gridView">подписываемый экземпляр типа GridView</param>
        /// <param name="view">Коллекция столбцов</param>
        private static void AddHandlers(GridView gridView, ICollectionView view)
        {
            GetGridViewsForColumnSource(view).Add(gridView);
            view.CollectionChanged += ColumnsSource_CollectionChanged;
        }
        /// <summary>
        /// Удаление и отписка от событий для переданной коллекции
        /// </summary>
        /// <param name="gridView">Отписываемый экземпляр типа GritView</param>
        /// <param name="view">Коллекция столбцов</param>
        private static void RemoveHandlers(GridView gridView, ICollectionView view)
        {
            view.CollectionChanged -= ColumnsSource_CollectionChanged;
            GetGridViewsForColumnSource(view).Remove(gridView);
        }
        /// <summary>
        /// Метод создающий колонки в подписываемом экземпляре класса GridView
        /// </summary>
        /// <param name="gridView">подписываемый экземпляр типа GridView</param>
        /// <param name="view">Коллекция столбцов</param>
        private static void CreateColumns(GridView gridView, IEnumerable<ColumnDescriptor> viewData)
        {
            foreach (var item in viewData)
            {
                GridViewColumn column = CreateColumn(gridView, item);
                int gv_count = gridView.Columns.Count(x => x.Header == column.Header);
                int v_count = viewData.Count(x => x.HeaderText.CompareTo(column.Header) == 0);
                if (gv_count < v_count)
                    gridView.Columns.Add(column);
            }
        }
        /// <summary>
        /// Коллбек события "CollectionChanged"
        /// </summary>
        /// <param name="sender">объект вызвавший событие</param>
        /// <param name="e">Данные об изменениях</param>
        private static void ColumnsSource_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            ICollectionView view = sender as ICollectionView;

            IEnumerable<ColumnDescriptor> viewData = view.Cast<ColumnDescriptor>();
            var gridViews = GetGridViewsForColumnSource(view);
            if (gridViews == null || gridViews.Count == 0)
                return;

            switch (e.Action)
            {
                // Обработка добавления новых столбцов
                case NotifyCollectionChangedAction.Add:
                    foreach (var gridView in gridViews)
                    {
                        for (int i = 0; i < e.NewItems.Count; i++)
                        {
                            GridViewColumn column = CreateColumn(gridView, e.NewItems[i]);
                            int gv_count = gridView.Columns.Count(x => x.Header == column.Header);
                            int v_count = viewData.Count(x => x.HeaderText.CompareTo(column.Header) == 0);
                            if (gv_count < v_count)
                                gridView.Columns.Insert(e.NewStartingIndex + i, column);
                        }
                    }
                    break;
                // Обработка перемещения столбцов внутри коллекции
                case NotifyCollectionChangedAction.Move:
                    foreach (var gridView in gridViews)
                    {
                        List<GridViewColumn> columns = new List<GridViewColumn>();
                        for (int i = 0; i < e.OldItems.Count; i++)
                        {
                            GridViewColumn column = gridView.Columns[e.OldStartingIndex + i];
                            columns.Add(column);
                        }
                        for (int i = 0; i < e.NewItems.Count; i++)
                        {
                            GridViewColumn column = columns[i];
                            gridView.Columns.Insert(e.NewStartingIndex + i, column);
                        }
                    }
                    break;
                // Обработка удаления столбцов из коллекции
                case NotifyCollectionChangedAction.Remove:
                    foreach (var gridView in gridViews)
                    {
                        for (int i = 0; i < e.OldItems.Count; i++)
                        {
                            gridView.Columns.RemoveAt(e.OldStartingIndex);
                        }
                    }
                    break;
                // Обработка замещения столбцов внутри коллекции
                case NotifyCollectionChangedAction.Replace:
                    foreach (var gridView in gridViews)
                    {
                        for (int i = 0; i < e.NewItems.Count; i++)
                        {
                            GridViewColumn column = CreateColumn(gridView, e.NewItems[i]);
                            gridView.Columns[e.NewStartingIndex + i] = column;
                        }
                    }
                    break;
                // Обработка полной отчистки коллекции
                case NotifyCollectionChangedAction.Reset:
                    foreach (var gridView in gridViews)
                    {
                        gridView.Columns.Clear();
                        CreateColumns(gridView, viewData);
                    }
                    break;
                default:
                    break;
            }
        }
        /// <summary>
        /// Создание колонки
        /// </summary>
        /// <param name="gridView">подписываемый экземпляр типа GridView</param>
        /// <param name="columnSource">объект описывающий колонку</param>
        /// <returns></returns>
        private static GridViewColumn CreateColumn(GridView gridView, object columnSource)
        {
            // Создание колонки
            GridViewColumn column = new GridViewColumn();
            // Получение имени заголовка таблицы и поля описывабщего колонку
            string headerTextMember = GetHeaderTextMember(gridView);
            string displayMemberMember = GetDisplayMemberMember(gridView);
            // Получение через рефлексию имен описывающих данные и Binding колонки с этими данными
            if (!string.IsNullOrEmpty(headerTextMember))
            {
                column.Header = GetPropertyValue(columnSource, headerTextMember);
            }
            if (!string.IsNullOrEmpty(displayMemberMember))
            {
                string propertyName = GetPropertyValue(columnSource, displayMemberMember) as string;
                column.DisplayMemberBinding = new Binding(propertyName);
            }
            return column;
        }
        /// <summary>
        /// Получение имен поля с которым осществляется связка колонки 
        /// а так же наименование заголовка столбца 
        /// с использованием механизма рефлексии
        /// </summary>
        /// <param name="obj">Экземпляр объекта описывающего колонку</param>
        /// <param name="propertyName">Имя поля в данном объекте которое хнарит информацию о колонки</param>
        /// <returns></returns>
        private static object GetPropertyValue(object obj, string propertyName)
        {
            // проверка передан ли объект
            if (obj != null)
            {
                // Получение информации о свойстве внутри объекта имя которому передано в переменной propertyName
                PropertyInfo prop = obj.GetType().GetProperty(propertyName);
                // Проверка получилось ли найти поле
                if (prop != null)
                    return prop.GetValue(obj, null); // Получение значения поля
            }
            return null;
        }
    }

    #region Accessory Objects
    /// <summary>
    /// Класс хранящий описание колонки
    /// </summary>
    class ColumnDescriptor
    {
        /// <summary>
        /// Описывает заголовок колонки
        /// </summary>
        public string HeaderText { get; set; }
        /// <summary>
        /// Описывает наименование эллемента с которым осуществляется связка данных
        /// </summary>
        public string DisplayMember { get; set; }
    }
    #endregion
}
